@@ -4,9 +4,9 @@ import hashlib
 import json
 import pathlib
 import re
-from typing import NamedTuple
-from threading import Thread
 from concurrent.futures.thread import ThreadPoolExecutor
+import dataclasses
+from threading import Thread
 
 import cachetools
 import requests
@@ -26,7 +26,7 @@ class Municipality:
     '''
     Provides info about all schools that use the Bakaláři system.\n
         >>> m = Municipality()
-        >>> for city in m.cities:
+        >>> for city in m.municipality().cities:
         >>>     print(city.name)
         >>>     for school in city.schools:
         >>>         print(school.name)
@@ -36,46 +36,57 @@ class Municipality:
                      Library comes prepackaged with the database json.
                      Use only when needed.
     '''
-    conf_dir = pathlib.Path(__file__).parent.joinpath("data")
-    db_file = conf_dir.joinpath("municipality.json")
+    data_dir = pathlib.Path(__file__).parent.joinpath("data")
+    db_file = data_dir.joinpath("municipality.json")
 
-    class City(NamedTuple):
+    @dataclasses.dataclass(frozen=True)
+    class Result:
+        cities: list
+
+    @dataclasses.dataclass(frozen=True)
+    class City:
         name: str
         school_count: str
         schools: list
 
-    class School(NamedTuple):
+    @dataclasses.dataclass(frozen=True)
+    class School:
         id: str
         name: str
         domain: str
 
     def __init__(self):
         super().__init__()
-        if not self.conf_dir.is_dir():
-            self.conf_dir.mkdir()
-        if self.db_file.is_file():
-            self.cities = [
-                self.City(
-                    city[0],
-                    city[1],
-                    [
-                        self.School(
-                            school[0],
-                            school[1],
-                            school[2]
-                        ) for school in city[2]
-                    ]
-                ) for city in json.loads(self.db_file.read_text(encoding='utf-8'), encoding='utf-8')
-            ]
-        else:
-            self.cities = self.build()
 
-    def build(self) -> dict:
+        self.thread = Thread(target=self._municipality)
+        self.thread.start()
+
+    def municipality(self):
+        if self.thread.is_alive():
+            self.thread.join()
+        return self._municipality()
+
+    def _municipality(self):
+        if not self.data_dir.is_dir():
+            self.data_dir.mkdir()
+        if self.db_file.is_file():
+            db = json.loads(self.db_file.read_text(encoding='utf-8'), encoding='utf-8')
+            result = self.Result([
+                self.City(name=city["name"], school_count=city["school_count"], schools=[
+                    self.School(id=school["id"], name=school["name"], domain=school["domain"]
+                    ) for school in city["schools"]
+                ]) for city in db["cities"]
+            ])
+            return result
+        else:
+            return self.build()
+
+    def build(self):
         import lxml.etree as ET
         url = "https://sluzby.bakalari.cz/api/v1/municipality/"
         parser = ET.XMLParser(recover=True)
 
-        cities = [
+        result = self.Result([
             self.City(
                 municInfo.find("name").text,
                 municInfo.find("schoolCount").text,
@@ -83,23 +94,27 @@ class Municipality:
                     self.School(
                         school.find("id").text,
                         school.find("name").text,
-                        re.sub("((/)?login.aspx(/)?)?", "", re.sub("http(s)?://(www.)?",
-                                                                   "", school.find("schoolUrl").text)).rstrip("/")
+                        re.sub("((/)?login.aspx(/)?)?", "",
+                            re.sub("http(s)?://(www.)?", "",
+                                school.find("schoolUrl").text)).rstrip("/")
                     ) for school in ET.fromstring(requests.get(url + requests.utils.quote(municInfo.find("name").text), stream=True).content, parser=parser).iter("schoolInfo") if school.find("name").text
                 ]
             ) for municInfo in ET.fromstring(requests.get(url, stream=True).content, parser=parser).iter("municipalityInfo") if municInfo.find("name").text
-        ]
+        ])
         self.db_file.write_text(json.dumps(
-            cities, indent=4, sort_keys=True), encoding='utf-8')
-        return cities
+            dataclasses.asdict(result), indent=4, sort_keys=True), encoding='utf-8')
+        return result
 
 
+cache = cachetools.TTLCache(32, 300)
+
+@cachetools.cached(cache)
 def request(url: str, token: str, *args) -> dict:
     '''
     Make a GET request to school URL.\n
     Module names are available at `https://github.com/bakalari-api/bakalari-api/tree/master/moduly`.
     '''
-    if args is None or len(args) > 2:
+    if not args or len(args) > 2:
         raise BakalibError("Bad arguments")
     params = {"hx": token, "pm": args[0]}
     if len(args) > 1:
@@ -114,42 +129,49 @@ def request(url: str, token: str, *args) -> dict:
     return response["results"]
 
 
-class Client(object):
+class Client:
     '''
-    Creates an instance with access to basic information of the user.
-    >>> user = Client(username="Username12345", domain="domain.example.com", password="1234abcd")
-    >>> user = Client(username="Username12345", domain="domain.example.com", perm_token="*login*Username12345*pwd*abcdefgh12345678+jklm==*sgn*ANDR")
+    Creates an instance with access to basic information of the user.\n
+    Check for validity runs only
+    >>> user = Client(username="User12345", domain="domain.example.com/bakaweb", password="1234abcd")
+    >>> user = Client(username="User12345", domain="domain.example.com/bakaweb", perm_token="*login*User12345*pwd*abcdefgh12345678+jklm==*sgn*ANDR")
+    >>> user.info()
     Methods:
         info(): Obtains basic information about the user.
         add_modules(*args): Extends the functionality with another module/s.
     '''
-    cache = cachetools.Cache(1)
-
-    def __init__(self, username: str, password=None, domain=None, perm_token=None):
+    def __init__(self, username: str, password: str=None, domain: str=None, perm_token: str=None, check_validity: bool=True):
         super().__init__()
         self.username = username
         self.domain = domain
         self.url = "https://{}/login.aspx".format(self.domain)
 
+        self.timetable = Timetable
+        self.grades = Grades
+
         if perm_token:
             self.perm_token = perm_token
-            self.token = self._token(self.perm_token)
+            token = self._token(self.perm_token)
         elif password:
             self.perm_token = self._permanent_token(username, password)
             token = self._token(self.perm_token)
-            if not self._is_token_valid(token):
-                raise BakalibError("Token is invalid\nInvalid password")
-            self.token = token
         else:
             raise BakalibError("Incorrect arguments")
 
-        self.thread = Thread(target=self._info)
-        self.thread.start()
+        if check_validity:
+            if self._is_token_valid(token):
+                self.token = token
+            else:
+                raise BakalibError("Token is invalid: Invalid password/perm_token")
+        else:
+            self.token = token
 
+        self.thread = Thread(target=request, args=(self.url, self.token, "login"))
+        self.thread.start()
+            
     def _permanent_token(self, user: str, password: str) -> str:
         '''
-        Generates a permanent access token with securely hashed password.\n
-        Returns a `str` containing the token.
+        Generates a permanent access token with securely hashed password.
         '''
         r = requests.get(url=self.url, params={"gethx": user}, verify=False)
         xml = xmltodict.parse(r.content)
@@ -166,6 +188,9 @@ class Client(object):
         return perm_token
 
     def _token(self, perm_token: str) -> str:
+        '''
+        Generates an access token using current time.
+        '''
         today = datetime.date.today()
         datecode = "{:04}{:02}{:02}".format(today.year, today.month, today.day)
         hash = hashlib.sha512((perm_token + datecode).encode("utf-8")).digest()
@@ -173,43 +198,49 @@ class Client(object):
         return token
 
     def _is_token_valid(self, token: str) -> bool:
+        '''
+        Checks for token validity.
+        '''
         try:
             request(self.url, token, "login")
             return True
         except BakalibError:
             return False
 
-    def add_modules(self, *modules):
+    def add_modules(self, *modules) -> None:
         '''
-        Extends the functionality of the Client class with another module(s).
+        Extends the functionality of the Client class with another module(s).\n
+        Supported modules: timetable, grades\n
+        WIP: absence
         >>> user.add_modules("timetable", "grades")
         >>> user.timetable.this_week()
+        >>> user.grades.grades()
         '''
         if modules:
             for module in modules:
                 if module == "timetable":
-                    self.timetable = Timetable(self.url, self.token)
+                    if self.timetable == Timetable:
+                        self.timetable = Timetable(self.url, self.token)
                 elif module == "grades":
-                    self.grades = Grades(self.url, self.token)
+                    if self.grades == Grades:
+                        self.grades = Grades(self.url, self.token)
                 else:
                     raise BakalibError("Bad module name was provided")
         else:
             raise BakalibError("No modules were provided")
 
     def info(self):
-        if self.thread.is_alive():
-            self.thread.join()
-        return self._info()
-
-    @cachetools.cached(cache)
-    def _info(self):
         '''
         Obtains basic information about the user into a NamedTuple.
         >>> user.info().name
-        >>> user.info().class_ # <-- due to class being reserved keyword.
+        >>> user.info().class_ # <-- due to class being a reserved keyword.
         >>> user.info().school
         '''
-        class Result(NamedTuple):
+        if self.thread.is_alive():
+            self.thread.join()
+
+        @dataclasses.dataclass(frozen=True)
+        class Result:
             version: str
             name: str
             type_abbr: str
@@ -231,23 +262,17 @@ class Client(object):
         )
         return result
 
-    def refresh_cache(self):
-        self.cache.clear()
-        self.thread.start()
 
-
-class Timetable(object):
+class Timetable:
     '''
     Obtains information from the "rozvrh" module of Bakaláři.
     >>> timetable = Timetable(url, token)
     Methods:
-        prev_week(): Decrements self.date by 7 days and points to date_week(self.date).
+        prev_week(): Decrements self.date by 7 days and points to self.date_week.
         this_week(): Points to date_week() with current date.
-        next_week(): Increments self.date by 7 days and points to date_week(self.date).
+        next_week(): Increments self.date by 7 days and points to self.date_week.
         date_week(date): Obtains all timetable data about the week of the provided date.
     '''
-    cache = cachetools.Cache(30)
-
     def __init__(self, url, token, date=datetime.date.today()):
         super().__init__()
         self.url = url
@@ -257,7 +282,7 @@ class Timetable(object):
         self.threadpool = ThreadPoolExecutor(max_workers=8)
         self.threadpool.submit(self._date_week, self.date)
 
-    # ------- convenience methods -------
+    # ----------------------------------------------------
 
     def prev_week(self):
         self.date = self.date - datetime.timedelta(7)
@@ -271,23 +296,9 @@ class Timetable(object):
         self.date = self.date + datetime.timedelta(7)
         return self.date_week(self.date)
 
-    # -----------------------------------
+    # ----------------------------------------------------
 
     def date_week(self, date=None):
-        self.date = date if date else self.date
-        if (self.date,) in self.cache:
-            return self._date_week(self.date)
-        else:
-            self.threadpool.shutdown(wait=True)
-            self.threadpool = ThreadPoolExecutor(max_workers=8)
-            self.threadpool.submit(
-                self._date_week, self.date - datetime.timedelta(7))
-            self.threadpool.submit(
-                self._date_week, self.date + datetime.timedelta(7))
-            return self._date_week(self.date)
-
-    @cachetools.cached(cache)
-    def _date_week(self, date):
         '''
         Obtains all timetable data about the week of the provided date.
         >>> this_week = timetable.date_week(datetime.date.today())
@@ -299,29 +310,51 @@ class Timetable(object):
         >>>         lesson.name
         >>>         lesson.teacher
         '''
+        global cache
+
+        self.date = date if date else self.date
+        date_str = "{:04}{:02}{:02}".format(date.year, date.month, date.day)
+
+        if not (self.url, self.token, "rozvrh", date_str) in cache:
+            self.threadpool.shutdown(wait=True)
+            self.threadpool = ThreadPoolExecutor(max_workers=8)
+
+        self.threadpool.submit(
+            self._date_week, self.date - datetime.timedelta(7))
+        self.threadpool.submit(
+            self._date_week, self.date + datetime.timedelta(7))
+        return self._date_week(self.date)
+
+    def _date_week(self, date):
+        date_str = "{:04}{:02}{:02}".format(date.year, date.month, date.day)
+
         response = request(
             self.url,
             self.token,
             "rozvrh",
-            "{:04}{:02}{:02}".format(date.year, date.month, date.day)
+            date_str
         )
 
-        class Result(NamedTuple):
+        @dataclasses.dataclass(frozen=True)
+        class Result:
             headers: list
             days: list
             cycle_name: str
 
-        class Header(NamedTuple):
+        @dataclasses.dataclass(frozen=True)
+        class Header:
             caption: str
             time_begin: str
             time_end: str
 
-        class Day(NamedTuple):
+        @dataclasses.dataclass(frozen=True)
+        class Day:
             abbr: str
             date: str
             lessons: list
 
-        class Lesson(NamedTuple):
+        @dataclasses.dataclass(frozen=True)
+        class Lesson:
             idcode: str
             type: str
             abbr: str
@@ -370,14 +403,8 @@ class Timetable(object):
         ]
         return Result(headers, days, response["rozvrh"]["nazevcyklu"])
 
-    def refresh_cache(self):
-        self.cache.clear()
-        self.threadpool.shutdown(wait=False)
-        self.threadpool = ThreadPoolExecutor(max_workers=8)
-        self.threadpool.submit(self._date_week, self.date)
 
-
-class Grades(object):
+class Grades:
     '''
     Obtains information from the "znamky" module of Bakaláři.
     >>> grades = Grades(url, token)
@@ -389,8 +416,6 @@ class Grades(object):
     Methods:
         grades(): Obtains all grades.
     '''
-    cache = cachetools.Cache(1)
-
     def __init__(self, url, token):
         super().__init__()
         self.url = url
@@ -412,7 +437,6 @@ class Grades(object):
             self.thread.join()
         return self._grades()
 
-    @cachetools.cached(cache)
     def _grades(self):
         response = request(
             self.url,
@@ -426,17 +450,20 @@ class Grades(object):
             if not isinstance(subject["znamky"]["znamka"], list):
                 response["predmety"]["predmet"][index]["znamky"]["znamka"] = [subject["znamky"]["znamka"]]
 
-        class Result(NamedTuple):
+        @dataclasses.dataclass(frozen=True)
+        class Result:
             subjects: list
 
-        class Subject(NamedTuple):
+        @dataclasses.dataclass(frozen=True)
+        class Subject:
             name: str
             abbr: str
             average_round: str
             average: str
             grades: list
 
-        class Grade(NamedTuple):
+        @dataclasses.dataclass(frozen=True)
+        class Grade:
             subject: str
             maxb: str
             grade: str
@@ -467,7 +494,7 @@ class Grades(object):
                         grade.get("caption"),
                         grade.get("typ"),
                         grade.get("ozn"),
-                        grade.get("poznámka")
+                        grade.get("poznamka")
                     ) for grade in subject["znamky"]["znamka"]
                 ]
             ) for subject in response["predmety"]["predmet"]
